/*
 * SPC input plugin for XMMS
 *
 * Copyright (C) 2000 AGAWA Koji <kaoru-k@self-core.104.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <xmms/plugin.h>
#include <xmms/configfile.h>
#include <xmms/util.h>
#include "spcxmms.h"
#include "libspc.h"

PluginConfig config;
static InputPlugin spc_ip;
static pthread_t decode_thread;
static gboolean audio_error = FALSE;
static gboolean going = FALSE;
static GtkWidget *w_about_window;
static gint mixbuf_size;
static gint finished_time;

/* ---------------------------------------------------------------- */
void config_set_default(PluginConfig *cfg)
{
    cfg->spc.sampling_rate = 44100;
    cfg->spc.resolution = 16;
    cfg->spc.channels = 2;
    cfg->spc.is_interpolation = TRUE;
    cfg->spc.is_echo = TRUE;
}

void config_read_from_default_file(PluginConfig *cfg)
{
    ConfigFile *cf;

    cf = xmms_cfg_open_default_file();

    xmms_cfg_read_int(cf, "SPC", "sampling_rate", &cfg->spc.sampling_rate);
    xmms_cfg_read_int(cf, "SPC", "resolution", &cfg->spc.resolution);
    xmms_cfg_read_int(cf, "SPC", "channels", &cfg->spc.channels);
    xmms_cfg_read_boolean(cf, "SPC", "is_interpolation", &cfg->spc.is_interpolation);
    xmms_cfg_read_boolean(cf, "SPC", "is_echo", &cfg->spc.is_echo);

    xmms_cfg_free(cf);
}

void config_write_to_default_file(PluginConfig *cfg)
{
    ConfigFile *cf;

    cf = xmms_cfg_open_default_file();

    xmms_cfg_write_int(cf, "SPC", "sampling_rate", cfg->spc.sampling_rate);
    xmms_cfg_write_int(cf, "SPC", "resolution", cfg->spc.resolution);
    xmms_cfg_write_int(cf, "SPC", "channels", cfg->spc.channels);
    xmms_cfg_write_boolean(cf, "SPC", "is_interpolation", cfg->spc.is_interpolation);
    xmms_cfg_write_boolean(cf, "SPC", "is_echo", cfg->spc.is_echo);

    xmms_cfg_write_default_file(cf);
    xmms_cfg_free(cf);
}  

/* ================================================================ */

static void initialize (void)
{
    config_set_default(&config);
    config_read_from_default_file(&config);
}

static int is_our_file (char *filename)
{
    gchar *ext;

    ext = strrchr(filename, '.');
    if (ext)
        if ((!strcasecmp(ext, ".spc")) || (!strcasecmp(ext, ".SPC")))
            return 1;

    return 0;
}

static void *play_loop(void *arg)
{
    gint8 *mixbuf;
    mixbuf = g_new(guint8, mixbuf_size);

    while (going) {
        SPC_update(mixbuf);
        spc_ip.add_vis_pcm(spc_ip.output->written_time(), FMT_S16_LE, 2,
                           mixbuf_size, mixbuf);

        while (spc_ip.output->buffer_free() < mixbuf_size && going)
            xmms_usleep(10000);

        if (going)
            spc_ip.output->write_audio(mixbuf, mixbuf_size);
    }

    /* Finish */
    g_free(mixbuf);
    finished_time = spc_ip.output->output_time();
    spc_ip.output->close_audio();
    pthread_exit(NULL);
}

static void play_file(char *filename)
{
    gchar *tmp, *name;

    /* initialize spc module */
    mixbuf_size = SPC_init(&config.spc);

    if (!SPC_load (filename))
        return;
  
    going = TRUE;
    audio_error = FALSE;

    spc_ip.output->open_audio ((config.spc.resolution == 16) ? FMT_S16_LE : FMT_U8,
                               config.spc.sampling_rate, config.spc.channels);
    if (spc_ip.output == NULL) {
        audio_error = TRUE;
        return;
    }

    tmp = strrchr(filename, '/');
    if (!tmp)
        tmp = filename;
    else
        tmp ++;
  
    name = g_strdup(tmp);
    *strrchr(name, '.') = '\0';
    spc_ip.set_info(name, -1, 16 * config.spc.sampling_rate, config.spc.sampling_rate,
                    config.spc.channels);
    g_free(name);

    pthread_create(&decode_thread, NULL, play_loop, NULL);
    return;
}

static void stop(void)
{
    if (going) {
        going = FALSE;
        pthread_join(decode_thread, NULL);
        SPC_close();
    }
}

static void ip_pause(short p)
{
    spc_ip.output->pause(p);
}

static void seek(int time)
{
    return;
}

static int get_time(void)
{
    if (audio_error)
        return -2;
    if (!going)
        return -1;
/*   if (!spc_ip.output->buffer_playing ()) */
/*     return -1; */

    return spc_ip.output->output_time();
}

static void get_song_info(char *filename, char **title, int *length)
{
    gchar *tmp, *name;

    tmp = strrchr(filename, '/');
    if (tmp == NULL)
        tmp = filename;
    else
        tmp ++;
  
    name = g_strdup(tmp);
    *strrchr(name, '.') = '\0';
    (*title) = name;
    (*length) = -1;
}

void on_config_changed(void)
{
    if (going) {
        going = FALSE;
        pthread_join(decode_thread, NULL);

        mixbuf_size = SPC_set_state(&config.spc);

        going = TRUE;
        spc_ip.output->open_audio((config.spc.resolution == 16) ? FMT_S16_LE : FMT_U8,
                                  config.spc.sampling_rate, config.spc.channels);
        if (spc_ip.output == NULL) {
            audio_error = TRUE;
            return;
	}
        spc_ip.output->flush(finished_time);

        pthread_create(&decode_thread, NULL, play_loop, NULL);
    } else
        mixbuf_size = SPC_set_state(&config.spc);
}

/* ---------------------------------------------------------------- */
static void about_cb(GtkButton *button, gpointer user_data)
{
    gtk_widget_destroy (w_about_window);
    w_about_window = NULL;
}

static void plugin_about(void)
{
    gchar *text =
        "spcxmms version " VERSION "\n"
        "SPC input plugin for XMMS\n\n"
        "XMMS interface written by AGAWA Koji <kaoru-k@self-core.org>\n"
        ;

    if (!w_about_window) {
        w_about_window = xmms_show_message("About SPC input plugin for XMMS ",
                                           text, "Close", TRUE,
                                           about_cb, NULL);
        gtk_widget_show(w_about_window);
    } else
        gdk_window_raise(w_about_window->window);
}

/* plugin information
   ---------------------------------------------------------------- */
void file_info_box(char *filename);
void configure(void);

static InputPlugin spc_ip = {
    NULL,
    NULL,
    "SPC Player " VERSION,	/* The description */
    initialize,			/* Called when the plugin is loaded */
    plugin_about,               /* Show the about box */
    configure,			/* configure */
    is_our_file,                /* Return 1 if the plugin can handle the file*/
    NULL,
    play_file,			/* Guess what... */
    stop,                       /* Tricky one */
    ip_pause,			/* Pause or unpause */
    seek,                       /* Seek to the specified time */
    NULL,                       /* Set the equalizer */
    get_time,			/* Get the time */
    NULL,                       /* Input-plugin specific volume functions */
    NULL,                       /* set_volume */
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    get_song_info,		/* Function to grab the title string */
    file_info_box,		/* Bring up an info window */
    NULL
};

InputPlugin *get_iplugin_info(void)
{
    return &spc_ip;
}
